#!/usr/bin/env python3
# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from omflut import FlutterGroup
from tacs.omflut.om import TacsBuilder
from omflut import RbfXferBuilder
from sdpm.api.om import SdpmBuilder
from pypk.api_om import PypkBuilder
import openmdao.api as om
import numpy as np

def get_path(fname):
    import os.path
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', fname)

def get_exp():
    mach = [0.499, 0.678, 0.901, 0.960] # experimental Mach number
    rho = [0.4278, 0.2082, 0.0995, 0.0634] # experimental density
    mu = [33.465, 68.753, 143.920, 225.820] # experimental mass ratio m/(rho*V)
    mach = [0.960] # experimental Mach number
    rho = [0.0634] # experimental density
    mu = [225.820] # experimental mass ratio m/(rho*V)
    n = len(mach)
    return n, mach, rho, mu

def get_cfg(irun):
    # Shared parameters
    c_ref = 0.559 # root chord
    k_ref = [0.01, 0.1, 0.2, 0.3] # reference reduced frequencies
    n_mod = 4 # number of modes
    # Experimental Mach number, density and mass ratio
    _, mach, rho, mu = get_exp()
    # Structural config
    s_cfg = tacs_cfg(n_mod)
    # SDPM config
    a_cfg = sdpm_cfg(c_ref, n_mod, mach[irun], k_ref)
    # RBF
    x_cfg = rbfi_cfg(n_mod)
    # NIPK
    f_cfg = pypk_cfg(c_ref, n_mod, mach[irun], rho[irun], mu[irun], k_ref)

    return s_cfg, a_cfg, x_cfg, f_cfg

def tacs_cfg(n_mod):
    from tacs import constitutive, elements
    # Material properties
    rho = 381.9  # density, kg/m^3
    e1 = 3.245e9 # Youngs' modulus in direction '1', Pa
    e2 = 4.162e8 # Youngs' modulus in direction '2', Pa
    nu12 = 0.31 # Poisson's ratio in plane '1-2'
    g12 = 4.392e8 # shear modulus in plane '1-2', Pa
    min_thickness = 0.0001
    max_thickness = 0.05

    # Element definition
    def element_callback(dvNum, compID, compDescript, elemDescripts, specialDVs, **kwargs):
        # Get the thickness of current element
        thickness = float(compDescript.split('/')[1].split('=')[1])
        # Set up orthotropic material and shell constitutive properties with 1 DV per element
        prop = constitutive.MaterialProperties(rho=rho, E1=e1, E2=e2, nu12=nu12, G12=g12)
        con = constitutive.IsoShellConstitutive(prop, t=thickness, tNum=dvNum, tlb=min_thickness, tub=max_thickness)
        # Add the element
        transform = elements.ShellRefAxisTransform(np.array([np.sqrt(2), np.sqrt(2), 0.0]))
        return elements.Quad4Shell(transform, con)

    # Various pyTACS options
    options = {
        'printtiming': True
    }

    return {
        'mesh_file': get_path('agard445.bdf'),
        'element_callback': element_callback,
        'num_modes': n_mod,
        'sigma': 1.,
        'pytacs_options': options,
        'write_solution': False
        }

def sdpm_cfg(c_ref, n_mod, mach, k_ref):
    wnc = 20
    return {
        # Verbosity
        'Verb_lvl': 0,
        # Model (geometry or mesh)
        'File': get_path('agard445.geo'), # input Gmsh file
        'Pars': {'nC': wnc, 'nS': 10, 'pC': 1.1, 'pS': 1.0}, # parameters for Gmsh file (optional)
        # Markers
        'Wing': 'wing', # name of the marker defining the lifting body (wing)
        'Te': 'te', # name of the marker defining the trailing edge of the lifting body
        # Freestream
        'Mach': mach, # Mach number (optional)
        'AoA': 0.0, # angle of attack (optional)
        'AoS': 0.0, # angle of side slip (optional)
        # Unsteady motion
        'Num_wake_div': wnc, # number of cells per chord length in the wake
        'Frequencies': k_ref, # list of reduced frequencies
        'Num_modes': n_mod, # number of modes
        # Geometry
        'Symmetry': True, # whether only half (symmetric) model is provided
        's_ref': 0.353, # reference surface area
        'c_ref': c_ref, # reference chord length
        'x_ref': 0.0, # reference point for moment computation (x)
        'y_ref': 0.0, # reference point for moment computation (y)
        'z_ref': 0.0 # reference point for moment computation (z)
    }

def rbfi_cfg(n_mod):
    return {
    'n_dim': 2, # number of coordinates to use for interpolation
    'n_modes': n_mod, # number of modes
    'smooth': 1 # smoothing
    }

def pypk_cfg(c_ref, n_mod, mach, rho, mu, k_ref):
    return {
        'name': 'agard', # case name
        'k_ref': np.array(k_ref), # reference reduced frequencies
        'l_ref': 0.5 * c_ref, # reference length (half root chord)
        'mach': mach, # freesteam Mach number
        'rho_inf': rho, # freesteam density
        'u_idx': np.linspace(0.2, 0.5, 31, endpoint=True), # velocity index range
        'mu': mu, # mass ratio
        'f_ref': 40.3511, # frequency of first torsional mode
        'n_modes': n_mod, # number of modes
        'rho_ks': 1e6, # KS aggregation parameter for KS
        'method': 'nipk', # method type
        'fluid': 'unmatched', # fluid type
        'vrb': 1 # verbosity level
    }

def mass_builder(mesh_file, element_callback):
    """Create TACS static problem to compute mass
    """
    from tacs.pytacs import pyTACS
    from tacs import functions
    from mpi4py import MPI
    fea_assembler = pyTACS(mesh_file, comm=MPI.COMM_WORLD)
    fea_assembler.initialize(element_callback)
    pbl = fea_assembler.createStaticProblem('static')
    pbl.addFunction('mass', functions.StructuralMass)
    return pbl

class AgardMass(om.ExplicitComponent):
    """Wrap TACS static problem to compute wing mass
    """
    def initialize(self):
        self.options.declare('pbl', desc='static problem', recordable=False)

    def setup(self):
        self.pbl = self.options['pbl']
        self.add_input('dv_struct', shape_by_conn=True, desc='structural design variables')
        self.add_output('m', val=0., desc='structural mass')
        self.declare_partials(of=['m'], wrt=['dv_struct'], method='exact')

    def compute(self, inputs, outputs):
        self.pbl.setDesignVars(inputs['dv_struct'])
        func = {}
        self.pbl.evalFunctions(func, ['mass'])
        outputs['m'] = func['static_mass']
        self.pbl.writeSolution()

    def compute_partials(self, inputs, partials):
        func_sens = {}
        self.pbl.evalFunctionsSens(func_sens, ['mass'])
        partials['m', 'dv_struct'] = func_sens['static_mass']['struct']

class Agard(om.Group):
    """Aerostructural flutter model of the AGARD 445.6 wing
    """
    def initialize(self):
        self.options.declare('n', default=1, desc='number of runs')

    def setup(self):
        self.n_run = self.options['n']
        self.add_subsystem('dvs', om.IndepVarComp(), promotes=['*'])
        for i in range(self.n_run):
            # Builders
            s_cfg, a_cfg, x_cfg, f_cfg = get_cfg(i)
            tacs = TacsBuilder(**s_cfg)
            sdpm = SdpmBuilder(a_cfg)
            rbfi = RbfXferBuilder(tacs, sdpm, x_cfg)
            pypk = PypkBuilder(f_cfg)
            # Flutter component
            self.add_subsystem(f'flutter_{i}', FlutterGroup(struct=tacs, xfer=rbfi, aero=sdpm, flutter=pypk))
        # Mass component
        pbl = mass_builder(s_cfg['mesh_file'], s_cfg['element_callback'])
        self.init_dvs = pbl.getDesignVars()
        self.add_subsystem('mass', AgardMass(pbl=pbl), promotes=['dv_struct'])

    def configure(self):
        # DV
        self.dvs.add_output('dv_struct', val=self.init_dvs, desc='structural DV')
        self.add_design_var('dv_struct', lower=0.0001, upper=0.05, scaler=1e3)
        for i in range(self.n_run):
            self.connect('dv_struct', f'flutter_{i}.struct.dv_struct')
        # CON
        for i in range(self.n_run):
            self.add_constraint(f'flutter_{i}.g_agg', upper=-0.001, scaler=1e3)
        # OBJ
        self.add_objective('mass.m', scaler=1)

def print_time(selapsed):
    days, rem = divmod(selapsed, 24*60*60)
    hours, rem = divmod(rem, 60*60)
    minutes, seconds = divmod(rem, 60)
    print('Wall-clock time: {:0>2}-{:0>2}:{:0>2}:{:0>2}'.format(int(days),int(hours),int(minutes),int(seconds)))

if __name__ == "__main__":
    import time
    # Set up the problem
    n_runs, mach, _, _ = get_exp()
    prob = om.Problem(model=Agard(n=n_runs), driver=om.ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-4, disp=True, maxiter=20))
    rec_inc = []
    for i in range(n_runs):
        rec_inc.extend([f'flutter_{i}.f', f'flutter_{i}.g', f'flutter_{i}.g_agg', f'flutter_{i}.flutter.f_speed', 'mass.m'])
    prob.model.recording_options['includes'] = rec_inc
    prob.model.add_recorder(om.SqliteRecorder('cases.sql'))
    prob.setup()
    om.n2(prob, show_browser=False, outfile='n2.html')

    # Run
    tic = time.perf_counter()
    #prob.run_model()
    #prob.check_partials(excludes=['*mass', '*aero', '*struct', '*xfer'], compact_print=False, form='central')
    #prob.check_totals(of=['flutter_0.g_agg'], wrt=['dv_struct'], compact_print=False, form='central', show_progress=True, driver_scaling=False)
    prob.run_driver()
    toc = time.perf_counter()
    print_time(toc-tic)

    # Print the history
    cases = om.CaseReader('cases.sql').get_cases()
    hdr = '{:>10s}'.format('it')
    for irun in range(n_runs):
        hdr += ', {:>10s}, {:>10s}, {:>10s}'.format(f'g_agg_{irun}', f'u_f_{irun}', 'm')
    print(hdr)
    uf = np.zeros(n_runs)
    for i, case in enumerate(cases):
        val = '{:10d}'.format(i)
        for irun in range(n_runs):
            g_agg = case.get_val(f'flutter_{irun}.g_agg')[0]
            uf[irun] = case.get_val(f'flutter_{irun}.flutter.f_speed')[0]
            m = case.get_val('mass.m')[0]
            val += ', {:10.4f}, {:10.4f}, {:10.4f}'.format(g_agg, uf[irun], m)
        print(val)
    # Save results
    np.savetxt('mach_uf.dat', np.column_stack((mach, uf)), delimiter=', ', fmt='%15.2f', header='{:>13s}, {:>15s}'.format('Mach number', 'Flutter speed'))
